@@ -1,4 +1,6 @@
 import inspect
+import subprocess
+import tempfile
 from copy import copy
 from weakref import WeakKeyDictionary
 
@@ -14,6 +16,7 @@ _keep_refs = WeakKeyDictionary()
 pyheif_supports_transformations = (
     'transformations' in inspect.signature(pyheif.HeifFile).parameters
 )
+HEIF_ENC_BIN = 'heif-enc'
 
 
 def _crop_heif_file(heif):
@@ -101,10 +104,6 @@ class HeifImageFile(ImageFile.ImageFile):
 
         _extract_heif_exif(heif_file)
 
-        if self._exclusive_fp:
-            self.fp.close()
-        self.fp = None
-
         if pyheif_supports_transformations:
             heif_file = _rotate_heif_file(heif_file)
             self._size = heif_file.transformations['crop'][2:4]
@@ -158,5 +157,73 @@ def check_heif_magic(data):
     return pyheif.check(data) != pyheif.heif_filetype_no
 
 
+def _save(im, fp, filename):
+    # Save it before subsequent im.save() call
+    info = im.encoderinfo
+
+    if im.mode in ('P', 'PA'):
+        # disbled due to errors in libheif encoder
+        raise IOError("cannot write mode P as HEIF")
+
+    with tempfile.NamedTemporaryFile(suffix='.png') as tmpfile:
+        if im.mode == '1':
+            # to circumvent `heif-enc` bug
+            im = im.convert('L')
+        im.save(
+            tmpfile, format='PNG', optimize=False, compress_level=0,
+            icc_profile=info.get('icc_profile', im.info.get('icc_profile')),
+            exif=info.get('exif', im.info.get('exif'))
+        )
+
+        cmd = [HEIF_ENC_BIN, '-o', '/dev/stdout', tmpfile.name]
+
+        avif = info.get('avif')
+        if avif is None and filename:
+            ext = filename.rpartition('.')[2].lower()
+            avif = ext == 'avif'
+        if avif:
+            cmd.append('-A')
+
+        if info.get('encoder'):
+            cmd.extend(['-e', info['encoder']])
+
+        if info.get('quality') is not None:
+            cmd.extend(['-q', str(info['quality'])])
+
+        subsampling = info.get('subsampling')
+        if subsampling is not None:
+            if subsampling == 0:
+                subsampling = '444'
+            elif subsampling == 1:
+                subsampling = '422'
+            elif subsampling == 2:
+                subsampling = '420'
+            cmd.extend(['-p', 'chroma=' + subsampling])
+
+        if info.get('speed') is not None:
+            cmd.extend(['-p', 'speed=' + str(info['speed'])])
+
+        if info.get('concurrency') is not None:
+            cmd.extend(['-p', 'threads=' + str(info['concurrency'])])
+
+        try:
+            # Warning: Do not open stdout and stderr at the same time
+            with subprocess.Popen(cmd, stdout=subprocess.PIPE) as enc:
+                for data in iter(lambda: enc.stdout.read(128 * 1024), b''):
+                    fp.write(data)
+                if enc.wait():
+                    raise subprocess.CalledProcessError(enc.returncode, cmd)
+        except FileNotFoundError:
+            raise FileNotFoundError(
+                2, f"Can't find heif encoding binary. Install '{HEIF_ENC_BIN}' "
+                + "or set `HeifImagePlugin.HEIF_ENC_BIN` to full path.")
+
+
 Image.register_open(HeifImageFile.format, HeifImageFile, check_heif_magic)
+Image.register_save(HeifImageFile.format, _save)
 Image.register_mime(HeifImageFile.format, 'image/heif')
+Image.register_extensions(HeifImageFile.format, [".heic", ".avif"])
+
+# Don't use this extensions for saving images, use the ones above.
+# They have added for quick file type detection only (i.g. by Django).
+Image.register_extensions(HeifImageFile.format, [".heif", ".hif"])
